@@ -1,4 +1,4 @@
-import fnmatch, traceback, sys, time, os.path, stat, pprint, re, signal, importlib
+import fnmatch, traceback, sys, time, os.path, stat, pprint, re, signal, importlib, platform
 
 from socket import *
 from pythreader import PyThread, synchronized, Task, TaskQueue, Primitive
@@ -10,8 +10,7 @@ from .logs import Logged, Logger
 from .py3 import PY2, PY3, to_str, to_bytes
 
 import pythreader
-        
-        
+
 class BodyFile(object):
     
     def __init__(self, buf, sock, length):
@@ -199,7 +198,7 @@ class HTTPHeader(object):
 class RequestProcessor(Task):
     
     def __init__(self, wsgi_app, request):
-        Task.__init__(self, name=f"[RequestTask {request.Id}]")
+        Task.__init__(self, name=f"[RequestProcessor {request.Id}]")
         #print("RequestTask: wsgi_app:", wsgi_app)
         self.WSGIApp = wsgi_app
         self.Request = request
@@ -225,24 +224,32 @@ class RequestProcessor(Task):
                 #print("env:")
                 #for k, v in env.items():
                 #    print(k,":",v)
-                out = self.WSGIApp(env, self.start_response)    
+                out = self.WSGIApp(env, self.start_response)
+                #print("HTTPServer: out:", out)
             except:
                 error = "error in wsgi_app: %s" % (traceback.format_exc(),)
                 csock.sendall(b"HTTP/1.1 500 Error\nContent-Type: text/plain\n\n"+to_bytes(error))
                 return self.error(error)
             
+
+            #print("RequestProcessor.run: out:", out)
             if self.OutBuffer:      # from start_response
+                #print("RequestProcessor.run: OutBuffer:", self.OutBuffer)
                 csock.sendall(to_bytes(self.OutBuffer))
                 
             self.ByteCount = 0
 
             for line in out:
                 line = to_bytes(line)
-                try:    csock.sendall(line)
+                try:
+                    #print("sending line:", line)    
+                    csock.sendall(line)
+                    #print("HTTPServer: sent:", line)
                 except Exception as e:
                     return self.error("error sending body: %s" % (e,))
                 self.ByteCount += len(line)
         finally:
+            #print("HTTPServer: closing request...")
             request.close()
             self.OutBuffer = None
             self.WSGIApp = None
@@ -262,17 +269,23 @@ class RequestProcessor(Task):
 
 class Service(Logged):
     
-    def __init__(self, app, logger=None):
+    def __init__(self, app, capacity=100, logger=None):
         Logged.__init__(self, f"[{app.__class__.__name__}]", logger=logger)
         self.Name = app.__class__.__name__
         self.WPApp = app
-        self.ProcessorQueue = TaskQueue(5, delegate=self)
+        self.ProcessorQueue = TaskQueue(5, capacity=capacity, delegate=self)
 
     def accept(self, request):
+        if not self.WPApp.match(request.HTTPHeader.URI):
+            return False, "no match"
         p = RequestProcessor(self.WPApp, request)
         request.AppName = self.Name
-        self.ProcessorQueue << p
-        return True
+        try:
+            self.ProcessorQueue.add(p, timeout=0)
+        except RuntimeError:
+            return True, "service unavailable"
+        else:
+            return True, "accepted"
 
     def taskFailed(self, queue, task, exc_type, exc_value, tb):
         self.error("request failed:", "".join(traceback.format_exception(exc_type, exc_value, tb)))
@@ -284,15 +297,14 @@ class Service(Logged):
     def taskEnded(self, queue, task, _):
         request = task.Request
         header = request.HTTPHeader
-        log_line = '%s:%s :%s %s %s -> %s %s %s %s' % (   
-                        request.CAddr[0], request.CAddr[1], request.ServerPort, 
+        log_line = '%s %s:%s :%s %s %s -> %s %s %s %s' % (   
+                        request.Id, request.CAddr[0], request.CAddr[1], request.ServerPort, 
                         header.Method, header.OriginalURI, 
                         request.AppName, header.path(),
                         task.StatusCode, task.ByteCount
                     )
         self.log(log_line)
 
-
 class Request(object):
     
     def __init__(self, port, csock, caddr):
@@ -308,8 +320,10 @@ class Request(object):
         
     def close(self):
         if self.CSock is not None:
-            try:    self.CSock.close()
-            except: pass
+            try:    
+                self.CSock.close()
+            except: 
+                pass
             self.CSock = None
         self.SSLInfo = None
 
@@ -393,6 +407,9 @@ class Request(object):
                     issuer = self.format_x509_name(x509.get_issuer())
         return subject, issuer
 
+    def send_response(self, status, headline):
+        response = f"HTTP/1.1 {status} {headline}\n\n"
+        self.CSock.sendall(to_bytes(response))
 
 class RequestReader(Task, Logged):
 
@@ -400,9 +417,8 @@ class RequestReader(Task, Logged):
 
     def __init__(self, dispatcher, request, socket_wrapper, timeout, logger):
         Task.__init__(self)
-        self.Request = request
-        csock_addr = request.CSock.getpeername()
-        Logged.__init__(self, f"[reader {request.Id} client:{csock_addr}]", logger=logger, debug=True)
+        self.Request = request       
+        Logged.__init__(self, f"[RequestReader {request.Id} client:%s:%s]" % request.CAddr, logger=logger, debug=True)
         self.SocketWrapper = socket_wrapper
         self.Dispatcher = dispatcher
         self.Timeout = timeout
@@ -423,6 +439,7 @@ class RequestReader(Task, Logged):
         csock = request.CSock
         saved_timeout = csock.gettimeout() 
         dispatched = False
+        dispatch_status = None
         try:
             #self.debug("started")
             self.Started = time.time()
@@ -453,16 +470,19 @@ class RequestReader(Task, Logged):
                     request.HTTPHeader = header
                     request.Body = body
                     self.debug("request received")
-                    service = self.Dispatcher.dispatch(self.Request)
-                    dispatched = service is not None
+                    dispatched, service, dispatch_status = self.Dispatcher.dispatch(self.Request)
         finally:
             if not dispatched:
                 if header is not None and header.Complete:
-                    try:    csock.sendall(b"HTTP/1.1 404 Not found\n\n")
-                    except: pass
-                    self.log('%s:%s :%s %s %s -> (nomatch)' % 
-                        (   request.CAddr[0], request.CAddr[1], request.ServerPort, 
-                            header.Method, header.OriginalURI
+                    if dispatch_status == "no match":
+                        request.send_response(404, "Service not found")
+                    elif dispatch_status == "service unavailable":
+                        request.send_response(503, "Service unavailable")
+                    else:
+                        request.send_response(500, "Request dispatch error " + dispatch_status)
+                    self.log('%s %s:%s :%s %s %s -> (%s)' % 
+                        (   request.Id, request.CAddr[0], request.CAddr[1], request.ServerPort, 
+                            header.Method, header.OriginalURI, dispatch_status
                         )
                     )
                 else:
@@ -507,7 +527,7 @@ class SSLSocketWrapper(object):
 
 class HTTPServer(PyThread, Logged):
 
-    def __init__(self, port, app=None, services=[], sock=None, logger=None, max_connections = 100, 
+    def __init__(self, port, app=None, services=[], sock=None, logger=None, max_connections = 100,
                 timeout = 20.0,
                 enabled = True, max_queued = 100,
                 logging = False, log_file = "-", debug=False,
@@ -525,12 +545,12 @@ class HTTPServer(PyThread, Logged):
         self.Timeout = timeout
         max_connections =  max_connections
         queue_capacity = max_queued
-        self.RequestReaderQueue = TaskQueue(max_connections, capacity=queue_capacity, delegate=self)
+        self.RequestReaderQueue = TaskQueue(max_connections, capacity=max_queued, delegate=self)
         self.SocketWrapper = SSLSocketWrapper(certfile, keyfile, verify, ca_file, password,
                 allow_proxies=allow_proxies) if keyfile else None
         
         if app is not None:
-            services = [Service(app, logger)]
+            services = [Service(app, logger=logger)]
             
         self.Services = services
         self.Stop = False
@@ -564,7 +584,8 @@ class HTTPServer(PyThread, Logged):
                 logging = logging, log_file=log_file, debug=debug,
                 certfile=certfile, keyfile=keyfile, verify=verify, ca_file=ca_file, password=password
         )
-        
+    
+    @synchronized
     def setServices(self, services):
         self.Services = services
         
@@ -611,13 +632,15 @@ class HTTPServer(PyThread, Logged):
         try:    self.Sock.close()
         except: pass
 
-    @synchronized
     def dispatch(self, request):
-        for service in self.Services:
-            if service.accept(request):
-                return service
+        with self:
+            services = self.Services[:]
+        for service in services:
+            match, status = service.accept(request)
+            if match:
+                return status == "accepted", service, status
         else:
-            return None
+            return False, None, "no match"
 
     def taskFailed(self, queue, task, exc_type, exc, tb):
         traceback.print_exception(exc_type, exc, tb)
