@@ -1,7 +1,9 @@
 from .webob import Response
+from .webob.multidict import MultiDict
 from .webob import Request as webob_request
-from .webob.exc import HTTPTemporaryRedirect, HTTPException, HTTPFound, HTTPForbidden, HTTPNotFound
+from .webob.exc import HTTPTemporaryRedirect, HTTPException, HTTPFound, HTTPForbidden, HTTPNotFound, HTTPBadRequest
 from . import Version as WebPieVersion
+from urllib.parse import unquote_plus
     
 import os.path, os, stat, sys, traceback, fnmatch, datetime, inspect, json
 from threading import RLock
@@ -19,7 +21,11 @@ else:
         return bytes(s)
     def to_str(b):    
         return str(b)
-    
+
+class InvalidArgumentError(Exception):
+    def __str__(self):
+        name, value = self.args
+        return f"Invalid argument value: {name}={value}"
 
 try:
     from collections.abc import Iterable    # Python3
@@ -41,10 +47,52 @@ _MIME_TYPES_BASE = {
         "css":  "text/css"
     }
 
+def _sql_quote_sanitizer(name, value):
+    return value.replace("'", "''")
+
+def _check_unsafe_sanitizer(name, value, unsafe="'"):
+    #print("_check_unsafe_sanitizer: name=", name, "   value:", value)
+    if value and any(c in value for c in unsafe):
+        raise InvalidArgumentError(name, value)
+    return value
+
+_Sanitizers = {
+    "sql":  _sql_quote_sanitizer,
+    "safe":  _check_unsafe_sanitizer
+}
+
 #
 # Decorators
 #
- 
+
+def sanitize(sanitizer=None, exclude=[], only=None):
+    def decorator(method):
+        szr = sanitizer
+        onl = only
+        excl = exclude
+        def decorated(handler, request, relpath, *params, **args):
+            "do-not-sanitize"       # signal "already sanitized"
+            sanitizer = szr
+            only = onl
+            exclude = excl
+            if isinstance(exclude, str):
+                exclude = [exclude]
+            if isinstance(only, str):
+                only = [only]
+            if sanitizer is None:
+                sanitizer = handler.App.Sanitizer
+            elif sanitizer is False:
+                sanitizer = None
+            elif isinstance(sanitizer, str):
+                sanitizer = handler.App.Sanitizers[sanitizer]
+            if sanitizer is not None:
+                #print("sanitizing...")
+                relpath, args = handler.App.sanitize(sanitizer, request, relpath, args, exclude=exclude, only=only)
+                #print("decorated: relpath:", relpath)
+            return method(handler, request, relpath, *params, **args)
+        return decorated
+    return decorator
+
 def webmethod(permissions=None):
     #
     # Usage:
@@ -85,7 +133,27 @@ def app_synchronized(method):
             return method(self, *params, **args)
     return synchronized_method
 
-atomic = app_synchronized
+atomic = app_synchronized           # synonym
+
+def _sql_quote_sanitizer(name, value):
+    return value.replace("'", "''")
+
+def _check_unsafe_sanitizer(name, value, unsafe="'"):
+    if value and any(c in value for c in unsafe):
+        raise UnsafeArgumentError(name, value)
+    return value
+
+def canonic_path(path):
+    # removes all occurances of '//' and '/./'
+    # if the path was absoulute, it remains absoulte (starts with '/')
+    # makes sure the path does not end with '/' unless it is the root path "/"
+    while path and '//' in path:
+        path = path.replace('//', '/')
+    while path and "/./" in path:
+        path = path.replace("/./","/")
+    if path and path != '/' and path.endswith('/'):
+        path = path[:-1]
+    return path
 
 class Request(webob_request):
     def __init__(self, *agrs, **kv):
@@ -112,11 +180,12 @@ class Request(webob_request):
         set_response_content_type,
         del_response_content_type, 
         "Response content type")
-
+        
 class HTTPResponseException(Exception):
     def __init__(self, response):
         self.value = response
 
+
 def makeResponse(resp):
     #
     # acceptable responses:
@@ -189,7 +258,6 @@ def makeResponse(resp):
     if content_type:
         response.content_type = content_type    # make sure to apply this after headers
     if text is not None:  response.text = text
-    #print(response)
     return response
 
 class WPHandler(object):
@@ -213,36 +281,6 @@ class WPHandler(object):
         if not self._Strict:
             self.addHandler(".env", self._env__)
             
-    def step_down(self, name):
-        if not name:    return self
-        allowed = not self._Strict
-        attr = None
-        if hasattr(self, name):
-            attr = getattr(self, name)
-        elif name in self._WebMethods:
-            attr = self._WebMethods[name]
-            allowed = True
-            
-        if attr is None:
-            return None
-            
-        if callable(attr):
-            allowed = allowed or (
-                        (self._MethodNames is not None 
-                                and name in self._MethodNames)
-                    or
-                        (hasattr(method, "__doc__") 
-                                and method.__doc__ == _WebMethodSignature)
-                    )
-            if not allowed:
-                return None
-            return attr
-        elif isinstance(attr, WPHandler):
-            return attr
-        else:
-            return None
-            
-        
     def addHandler(self, name, method):
         self._WebMethods[name] = method
 
@@ -258,7 +296,7 @@ class WPHandler(object):
             #self.apacheLog("roles: %s" % (roles,))
             return self.checkRoles(roles)
         return True
-        
+
     def checkRoles(self, roles):
         # override me
         return True
@@ -276,7 +314,7 @@ class WPHandler(object):
         self.BeingDestroyed = False
         
     def canonicPath(self, path):
-        return self.App.canonicPath(path)
+        return canonic_path(path)
 
     def externalPath(self, path):
         return self.App.externalPath(path)
@@ -321,6 +359,18 @@ class WPHandler(object):
             buf.append(l)
         if buf:
             yield ''.join(buf)
+            
+    def query_string(self, args):
+        parts = []
+        for name, values in args.items():
+            if not isinstance(values, list):
+                values = [values]
+            for v in values:
+                if v is None:
+                    parts.append(name)
+                else:
+                    parts.append(f"{name}={v}")
+        return "&".join(parts)
 
     def render_to_response_iterator(self, temp, _merge_lines=0,
                     **more_args):
@@ -369,19 +419,75 @@ class WPHandler(object):
         
     #
     # This web methods can be used for debugging
-    # call it as "../.env"
+    # call it as ".../.env"
     #
 
     def _env__(self, req, relpath, **args):
         lines = (
             ["request.environ:"]
             + ["  %s = %s" % (k, repr(v)) for k, v in sorted(req.environ.items())]
-            + ["relpath: %s" % (relpath or "")]
-            + ["args:"]
+            + ["\nrelpath: %s" % (relpath or "")]
+            + ["\nargs:"]
             + ["  %s = %s" % (k, repr(v)) for k, v in args.items()]
+            + ["\nrequest.GET:"]
+            + ["  %s = %s" % (k, repr(v)) for k, v in req.GET.items()]
+            + ["\nrequest.POST:"]
+            + ["  %s = %s" % (k, repr(v)) for k, v in req.POST.items()]
         )
         return "\n".join(lines) + "\n", "text/plain"
+
+    def _handle_request(self, request, path, path_down, args):
+        orig_path = canonic_path("/".join([path]+path_down))
+        word = ""
+        while path_down and not word:
+            word, path_down = path_down[0], path_down[1:]
+        relpath = "/".join(path_down)
+
+        if not word:
+            if self.DefaultMethod:
+                qs = self.query_string(args)
+                uri = "./" + self.DefaultMethod
+                if qs:
+                    uri += "?" + uri
+                self.redirect(uri)
+            else:
+                raise HTTPNotFound("no default method defined")
+
+        self.Path = path or "/"
+        subhandler = None
+        allowed = not self._Strict
         
+        if hasattr(self, word):
+            subhandler = getattr(self, word)
+        elif word in self._WebMethods:
+            subhandler = self._WebMethods[word]
+            allowed = True
+
+        if subhandler is None:
+            raise HTTPNotFound("invalid path: " + orig_path)
+
+        if isinstance(subhandler, Response):
+            return subhandler
+        if isinstance(subhandler, tuple):
+            return makeResponse(subhandler)
+        elif callable(subhandler):
+            allowed = allowed and not word.startswith('_')
+            allowed = allowed or (
+                        (self._MethodNames is not None 
+                                and word in self._MethodNames)
+                    or
+                        (hasattr(subhandler, "__doc__") 
+                                and subhandler.__doc__ == _WebMethodSignature)
+            )
+            if not allowed:
+                raise HTTPNotFound("invalid path: " + orig_path)
+            else:
+                return subhandler(request, relpath, **args)
+        elif isinstance(subhandler, WPHandler):
+            return subhandler._handle_request(request, path + "/" + word, path_down, args)
+        else:
+            raise HTTPNotFound("invalid path: " + orig_path)
+
 class WPStaticHandler(WPHandler):
     
     def __init__(self, request, app, root="static", default_file="index.html", cache_ttl=None):
@@ -446,8 +552,10 @@ class WPStaticHandler(WPHandler):
 class WPApp(object):
 
     Version = "Undefined"
+    Sanitizers = _Sanitizers
 
-    def __init__(self, root_class_or_handler, strict=False, prefix=None, replace_prefix="", environ={}):
+    def __init__(self, root_class_or_handler, strict=False, prefix=None, replace_prefix="", 
+            environ={}, unquote_args=True, sanitizer=None):
 
         self.RootHandler = self.RootClass = None
         if inspect.isclass(root_class_or_handler):
@@ -464,16 +572,24 @@ class WPApp(object):
         self.HandlerParams = []
         self.HandlerArgs = {}
         self.Environ = environ
+        self.UnquoteArgs = unquote_args
+
+        if isinstance(sanitizer, str):
+            sanitizer = self.Sanitizers[sanitizer]
+        self.Sanitizer = sanitizer
         
+    def match(self, uri):
+        return not self.Prefix or uri.startswith(self.Prefix)
+
     def _app_lock(self):
         return self._AppLock
-        
+
     def __enter__(self):
         return self._AppLock.__enter__()
-        
+
     def __exit__(self, *params):
         return self._AppLock.__exit__(*params)
-    
+
     # override
     @app_synchronized
     def acceptIncomingTransfer(self, method, uri, headers):
@@ -482,18 +598,6 @@ class WPApp(object):
     def init(self):
         pass
 
-    def canonicPath(self, path):
-        # removes all occurances of '//' and '/./'
-        # if the path was absoulute, it remains absoulte (starts with '/')
-        # makes sure the path does not end with '/' unless it is the root path "/"
-        while path and '//' in path:
-            path = path.replace('//', '/')
-        while path and "/./" in path:
-            path = path.replace("/./","/")
-        if path and path != '/' and path.endswith('/'):
-            path = path[:-1]
-        return path
-
     @app_synchronized
     def initJinjaEnvironment(self, tempdirs = [], filters = {}, globals = {}):
         # to be called by subclass
@@ -523,7 +627,7 @@ class WPApp(object):
         exc_text = ''.join(exc_text)
         text = """<html><body><h2>Application error</h2>
             <h3>%s</h3>
-            <pre>%s</pre>
+            <pre>\n%s</pre>
             </body>
             </html>""" % (headline, exc_text)
         #print exc_text
@@ -539,56 +643,22 @@ class WPApp(object):
                 
             if matched is None:
                 return None
-                
+
+            path = path[len(matched):]
+
             if self.ReplacePrefix:
-                path = self.ReplacePrefix + path[len(matched):]
+                path = self.ReplacePrefix + path
                 
-            path = path or "/"
+            path = canonic_path(path or "/")
             #print(f"converted to: [{path}]")
-                
+
         return path
-        
+
     def handler_options(self, *params, **args):
         self.HandlerParams = params
         self.HandlerArgs = args
         return self
 
-    def find_web_method(self, handler, request, path, path_down, args):
-        #
-        # walks down the tree of handler finds the web method and calls it
-        # returs the Response
-        #
-        
-        path = path or "/"
-        method = None
-        while path_down and not path_down[0]:
-            path_down.pop(0)
-            
-        #is_wp_handler = isinstance(handler, WPHandler)
-        #print(f"find_web_method({handler}, WPHandler:{is_wp_handler}, path={path}, path_down={path_down})")
-
-        if isinstance(handler, WPHandler):  
-            handler.Path = path
-            
-            if path_down:
-                name = path_down[0]
-                attr = handler.step_down(name)
-                if attr is not None:
-                    if not path.endswith("/"):  path += "/"
-                    return self.find_web_method(attr, request, path + name, path_down[1:], args)
-                    
-            if callable(handler):
-                method = handler
-            elif not path_down:
-                prefix = (path.split("/")[-1] or ".") + "/"
-                redirect = prefix + handler.DefaultMethod
-                raise HTTPFound(location=redirect)
-        elif callable(handler):
-            method = handler
-            
-        relpath = "/".join(path_down)
-        return method, relpath
-
     def parseQuery(self, query):
         out = {}
         for w in (query or "").split("&"):
@@ -597,48 +667,70 @@ class WPApp(object):
                 k = words[0]
                 if k:
                     v = None
-                    if len(words) > 1:  v = words[1]
+                    if len(words) > 1:  v = unquote_plus(words[1]) if self.UnquoteArgs else words[1]
                     if k in out:
                         old = out[k]
-                        if type(old) != type([]):
-                            old = [old]
-                            out[k] = old
-                        out[k].append(v)
+                        if not isinstance(old, list):
+                            old = out[k] = [old]
+                        old.append(v)
                     else:
                         out[k] = v
         return out
-        
+
+    def sanitize(self, sanitizer, request, relpath, args, exclude=[], only=None):
+        if sanitizer is None:
+            return relpath, args
+        if "(relpath)" not in exclude:  
+            relpath = sanitizer('(relpath)', relpath)
+        args = {
+                name: value if value is None 
+                        or name in exclude 
+                        or only is not None and name not in only
+                else (
+                    [sanitizer(name, v) for v in value] 
+                    if isinstance(value, list)
+                    else sanitizer(name, value)
+                ) 
+                for name, value in args.items()
+        }
+        #
+        # run sanitizer on GET and POST dictionaries of the request too, even though
+        # the values will not be updated. But at least the sanitizer has a chance to raise the UnsafeArgumentError exception
+        #
+        for name, value in list(request.GET.items()) + list(request.POST.items()):
+            if name not in exclude and (only is None or name in olny):
+                sanitizer(name, value)
+        return relpath, args
+
     def wsgi_call(self, root_handler, environ, start_response):
-        # path_to = '/'
-        path = environ.get('PATH_INFO', '')
+        path = canonic_path(environ.get('PATH_INFO', ''))
         #while "//" in path:
         #    path.replace("//", "/")
         path_down = path.split("/")
+        if not path_down[0]:
+            path_down = path_down[1:]
         #while '' in path_down:
         #    path_down.remove('')
         args = self.parseQuery(environ.get("QUERY_STRING", ""))
         request = Request(environ)
         try:
-            method, relpath = self.find_web_method(root_handler, request, "", path_down, args)
-            #print("WPApp.wsgi_call: method:", method, "   relpath:", relpath)
-            if method is None:
-                response = HTTPNotFound("Invalid path %s" % (path,))
+            if isinstance(root_handler, tuple):
+                response = makeResponse(root_handler)
+            elif isinstance(root_handler, Response):
+                response = root_handler
+            elif callable(root_handler):
+                response = root_handler(request, path, **args)
             else:
-                #print("method:", method)
-                response = method(request, relpath, **args)  
-                #print("response:", response)                  
-            
-        except HTTPFound as val:    
-            # redirect
-            response = val
+                response = root_handler._handle_request(request, "", path_down, args)
         except HTTPException as val:
-            #print 'caught:', type(val), val
             response = val
         except HTTPResponseException as val:
             #print 'caught:', type(val), val
             response = val
-        except:
-            response = self.applicationErrorResponse("Uncaught exception", sys.exc_info())
+        except InvalidArgumentError as e:
+            response = HTTPBadRequest(str(e))
+        except Exception as e:
+            response = self.applicationErrorResponse(str(e), sys.exc_info())
 
         try:    
             response = makeResponse(response)
@@ -666,7 +758,7 @@ class WPApp(object):
         assert path.startswith("/"), f"Can not convert relative path {path} to external path"
         if self.ReplacePrefix and path.startswith(self.ReplacePrefix):
             path = path[len(self.ReplacePrefix):]
-        return self.canonicPath(self.ExternalAppRootPath + '/' + path)
+        return canonic_path(self.ExternalAppRootPath + '/' + path)
 
     def __call__(self, environ, start_response):
         path = environ.get('PATH_INFO', '')
@@ -692,7 +784,7 @@ class WPApp(object):
                 self.ScriptName = environ.get('SCRIPT_NAME') or os.environ.get('SCRIPT_NAME', '')
                 self.Script = environ.get('SCRIPT_FILENAME') or os.environ.get('UWSGI_SCRIPT_FILENAME')
                 self.ScriptHome = os.environ.get('WEBPIE_SCRIPT_HOME') or os.path.dirname(self.Script or sys.argv[0]) or "."
-                self.ExternalAppRootPath = self.canonicPath('/' + self.ScriptName + '/' + (self.Prefix or ""))
+                self.ExternalAppRootPath = canonic_path('/' + self.ScriptName + '/' + (self.Prefix or ""))
                 self.init()
                 self.Initialized = True
 
@@ -705,7 +797,8 @@ class WPApp(object):
         #print("root_handler:", root_handler)
             
         try:
-            return self.wsgi_call(root_handler, environ, start_response)
+            out = self.wsgi_call(root_handler, environ, start_response)
+            return out
         except:
             resp = self.applicationErrorResponse(
                 "Uncaught exception", sys.exc_info())
@@ -780,4 +873,4 @@ if __name__ == '__main__':
     class MyHandler(WPHandler):
         pass
             
-    MyApp(MyHandler).run_server(8080)
\ No newline at end of file
+    MyApp(MyHandler).run_server(8080)
