@@ -4,26 +4,23 @@ import string, random
 
 class _UIDGen(Primitive):
     
-    def __init__(self, start=1, multiplier=1, offset=0, tag=""):
+    def __init__(self, tag="", range=1000000):
         Primitive.__init__(self)
-        self.Range = 1000000
-        self.Next = start
+        self.Next = 0
         self.Tag = tag
-        self.Offset = offset
-        self.Multiplier = multiplier
+        self.Range = range
 
     _alphabet=string.ascii_lowercase + string.ascii_uppercase
 
     @synchronized
     def get(self, as_int=False):
-        u = self.Next
         self.Next = (self.Next + 1) % self.Range
-        u = self.Next * self.Multiplier + self.Offset
+        u = self.Next
         if not as_int:
             a1 = random.choice(self._alphabet)
             a2 = random.choice(self._alphabet)
             a3 = random.choice(self._alphabet)
-            u = "%03d.%s%s%s.%03d" % (u//1000, a1, a2, a3, u%1000)
+            u = "%s%s%s.%03d" % (a1, a2, a3, u%1000)
             if self.Tag:
                 u = self.Tag + "." + u
         return u
@@ -38,8 +35,8 @@ def uid(u=None, as_int=False, tag=""):
     u = _uid.get(as_int)
     return u
     
-def init(multiplier=1, tag="", offset=0, start=1):
+def init(tag=""):
     global _uid
-    _uid = _UIDGen(start, multiplier, offset, tag)
+    _uid = _UIDGen(tag)
     
     
